import os
import wget
import numpy as np
import healpy as hp
from pathlib import Path
import matplotlib.pyplot as plt
from astropy.stats import median_absolute_deviation
from itertools import cycle

from matplotlib import rcParams
from os import environ
import scipy.optimize as opt
from mpl_toolkits.axes_grid1 import make_axes_locatable

import mwa_pb
fee_dir = Path(f'{os.path.dirname(mwa_pb.__file__)}/data')
if Path(f'{fee_dir}/mwa_full_embedded_element_pattern.h5').is_file() is not True:
    print('Downloading MWA FEE model from Cerberus')
    url = 'http://cerberus.mwa128t.org/mwa_full_embedded_element_pattern.h5'
    wget.download(url, f'{fee_dir}')

MWAPY_H5PATH=f'{fee_dir}/mwa_full_embedded_element_pattern.h5'
from mwa_pb import primary_beam
from mwa_pb import beam_full_EE
from mwa_pb import mwa_tile
from mwa_pb.mwa_sweet_spots import all_grid_points
from plot_tile_maps import plot_healpix


def local_beam(za, az, freq, delays=None, zenithnorm=True, power=True, jones=False, interp=True, pixels_per_deg=5,amps=None):
    '''Code pulled from mwapy that generates the MWA beam response - removes unecessary extra code from mwapy/pb
        - delays is a 2x16 array, with the first 16 delays for the XX, 
          second 16 for the YY pol. values match what you find in the metafits file
        - amps are the amplitudes of each individual dipole, again in a 2x16,
            with XX first then YY'''

    tile=beam_full_EE.ApertureArray(MWAPY_H5PATH,freq)
    mybeam=beam_full_EE.Beam(tile, delays,amps)
    if interp:
        j=mybeam.get_interp_response(az, za, pixels_per_deg)
    else:
        j=mybeam.get_response(az, za)
    if zenithnorm==True:
        j=tile.apply_zenith_norm_Jones(j) #Normalise

    #Use swapaxis to place jones matrices in last 2 dimensions
    #insead of first 2 dims.
    if len(j.shape)==4:
        j=np.swapaxes(np.swapaxes(j,0,2),1,3)
    elif len(j.shape)==3: #1-D
        j=np.swapaxes(np.swapaxes(j,1,2),0,1)
    else: #single value
        pass

    if jones:
        return j

    #Use mwa_tile makeUnpolInstrumentalResponse because we have swapped axes
    vis = mwa_tile.makeUnpolInstrumentalResponse(j,j)
    if not power:
        return (np.sqrt(vis[:,:,0,0].real),np.sqrt(vis[:,:,1,1].real))
    else:
        return (vis[:,:,0,0].real,vis[:,:,1,1].real)

   


if __name__=='__main__':

    import sys
    import argparse
    from pathlib import Path
    sys.path.append('../decode_rf_data')
    from colormap import jade
    import matplotlib.pyplot as plt
    
    # Custom spectral colormap
    jade, _ = jade()

    parser = argparse.ArgumentParser(description="""
        Create Simulated MWA Beam response maps using mwapy
        """)
    
    parser.add_argument('--out_dir', metavar='\b', default='./../../outputs/tile_maps/FEE_maps/', help='Output directory. Default=./../../outputs/tile_maps/FEE_maps/')
    parser.add_argument('--nside', metavar='\b', type=int,  default=32,help='Healpix Nside. Default = 32')
    
    args = parser.parse_args()
    
    out_dir     = Path(args.out_dir)
    nside       = args.nside
    
    # make output directory if it doesn't exist
    Path(out_dir).mkdir(parents=True, exist_ok=True)


    pointings = ['0', '2', '4', '41']

    
    fee_beam = {}
    for p in pointings:

        # Empty array for model beam
        npix = hp.nside2npix(nside)
        beam_response_XX = np.zeros(npix)
        beam_response_YY = np.zeros(npix)

        # healpix indices above horizon
        # convert to zenith angle and azimuth
        above_horizon = range(int(npix/2))
        beam_zas,beam_azs = hp.pix2ang(nside, above_horizon)
   
        delay_p = np.array([all_grid_points[int(p)][-1], all_grid_points[int(p)][-1]])

        # S21 had a missing dipole, so need a different amplitude array for the model
        amps = np.ones((2,16))
        #if AUT == 'S21XX':
        #    amps[0,5] = 0

        # Make beam response
        response = local_beam([list(beam_zas)], [list(beam_azs)], freq=137e+6, delays=delay_p, zenithnorm=True, power=True, interp=False, amps=amps)
        response_XX = response[0][0]
        response_YY = response[1][0]
        
        # Stick in an array, convert to decibels, and noralise
        beam_response_XX[above_horizon] = response_XX
        decibel_beam_XX = 10*np.log10(beam_response_XX)
        normed_beam_XX = decibel_beam_XX - decibel_beam_XX.max()
        
        beam_response_YY[above_horizon] = response_YY
        decibel_beam_YY = 10*np.log10(beam_response_YY)
        normed_beam_YY = decibel_beam_YY - decibel_beam_YY.max()
        
        fee_beam[p] = [normed_beam_XX, normed_beam_YY]
      
        fig = plt.figure(figsize=(9,10))
        fig.suptitle(f'MWA FEE MAP @ pointing [{p}] XX', fontsize=16, y=1.0)
        plot_healpix(data_map=normed_beam_XX, sub=(1,1,1), cmap=jade, vmin=-50, vmax=0)
        plt.savefig(f'{out_dir}/mwa_fee_beam_{p}_XX.png')
        plt.close()
        
        fig = plt.figure(figsize=(9,10))
        fig.suptitle(f'MWA FEE MAP @ pointing [{p}] YY', fontsize=16, y=1.0)
        plot_healpix(data_map=normed_beam_YY, sub=(1,1,1), cmap=jade, vmin=-50, vmax=0)
        plt.savefig(f'{out_dir}/mwa_fee_beam_{p}_YY.png')
        plt.close()

    np.savez_compressed(f'{out_dir}/mwa_fee_beam.npz', **fee_beam)


    fee_beam_flagged = {}
    for p in pointings:

        # Empty array for model beam
        npix = hp.nside2npix(nside)
        beam_response_XX = np.zeros(npix)
        beam_response_YY = np.zeros(npix)

        # healpix indices above horizon
        # convert to zenith angle and azimuth
        above_horizon = range(int(npix/2))
        beam_zas,beam_azs = hp.pix2ang(nside, above_horizon)
   
        delay_p = np.array([all_grid_points[int(p)][-1], all_grid_points[int(p)][-1]])

        # S21 had a missing dipole, so need a different amplitude array for the model
        amps = np.ones((2,16))
        amps[1,8] = 0

        # Make beam response
        response = local_beam([list(beam_zas)], [list(beam_azs)], freq=137e+6, delays=delay_p, zenithnorm=True, power=True, interp=False, amps=amps)
        response_XX = response[0][0]
        response_YY = response[1][0]
        
        # Stick in an array, convert to decibels, and noralise
        beam_response_XX[above_horizon] = response_XX
        decibel_beam_XX = 10*np.log10(beam_response_XX)
        normed_beam_XX = decibel_beam_XX - decibel_beam_XX.max()
        
        beam_response_YY[above_horizon] = response_YY
        decibel_beam_YY = 10*np.log10(beam_response_YY)
        normed_beam_YY = decibel_beam_YY - decibel_beam_YY.max()
        
        fee_beam_flagged[p] = [normed_beam_XX, normed_beam_YY]
      
        #fig = plt.figure(figsize=(9,10))
        #fig.suptitle(f'MWA FEE MAP @ pointing [{p}] XX', fontsize=16, y=1.0)
        #plot_healpix(data_map=normed_beam_XX, sub=(1,1,1), cmap=jade, vmin=-50, vmax=0)
        #plt.savefig(f'{out_dir}/mwa_fee_beam_{p}_XX.png')
        #plt.close()
        
        fig = plt.figure(figsize=(9,10))
        fig.suptitle(f'MWA FEE MAP @ pointing [{p}] YY', fontsize=16, y=1.0)
        plot_healpix(data_map=normed_beam_YY, sub=(1,1,1), cmap=jade, vmin=-50, vmax=0)
        plt.savefig(f'{out_dir}/mwa_fee_beam_{p}_YY_9_flagged.png')
        plt.close()

    np.savez_compressed(f'{out_dir}/mwa_fee_beam_9_flagged.npz', **fee_beam_flagged)



# Code developed by Jack Line and adapted here
# https://github.com/JLBLine/MWA_ORBCOMM

import numpy as np
import healpy as hp
from pathlib import Path
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from scipy import interpolate
from scipy.interpolate import RectSphereBivariateSpline,SmoothSphereBivariateSpline

import sys
sys.path.append('../decode_rf_data')
from colormap import spectral
cmap = spectral()
    
def create_model(len_empty_healpix, epsilon, nside, file_name=None):
    '''Takes .ffe model, converts into healpix and smooths the response'''

    #Make an empty array for healpix projection
    beam_response = np.zeros(len_empty_healpix)*np.nan

    # Load ffe model data in, which is stored in real and imaginary components
    # of a phi/theta polaristaion representation of the beam
    # apparently this is normal to beam engineers
    data = np.loadtxt(file_name)
    theta = data[:,0]
    phi = data[:,1]
    re_theta = data[:,2]
    im_theta = data[:,3]
    re_phi = data[:,4]
    im_phi = data[:,5]

    # Convert to complex numbers
    X_theta = re_theta.astype(complex)
    X_theta.imag = im_theta

    X_phi = re_phi.astype(complex)
    X_phi.imag = im_phi


    # make a coord grid for the data in the .ffe model
    theta_range = np.linspace(epsilon,90,91)
    phi_range = np.linspace(epsilon,359.0,360)
    theta_mesh,phi_mesh = np.meshgrid(theta_range,phi_range)

    # Convert reference model into power from the complex values
    power = 10*np.log10(abs(X_theta)**2 + abs(X_phi)**2)
    
    # Make an inf values super small
    power[np.where(power == -np.inf)] = -80

    # Had a problem with edge effects, leave off near horizon values
    # Basically remove the last 90 values of power list, where θ == 90
    power = power[:-91]

    # Get things into correct shape and do an interpolation
    power.shape = phi_mesh.shape
    
    # Bivariate spline approximation over a rectangular mesh on a sphere
    # s is a paramater I had to play with to get by eye nice results
    # s: positive smoothing factor
    lut = RectSphereBivariateSpline(theta_range*(np.pi/180.0), phi_range*(np.pi/180.0), power.T,s=0.1)

    # Get the theta and phi of all healpixels
    i_pix = np.arange(hp.nside2npix(nside))
    theta_rad, phi_rad = hp.pix2ang(nside, i_pix)

    # Evaluate the interpolated function at healpix gridpoints
    # Use spline to map beam into healpixels
    beam_response = lut.ev(theta_rad, phi_rad)
    
    # rescale beam response to have peak value of 0 dB
    beam_response = beam_response - np.nanmax(beam_response)

    return beam_response,theta_mesh,phi_mesh,power,theta


def plot_healpix(data_map=None, fig=None, sub=None,title=None,vmin=None,vmax=None,cmap=None, cbar=True):
    '''Yeesh do some healpix magic to plot the thing'''
    
    # Disable cryptic healpy warnings. Can't figure out where they originate
    import warnings
    warnings.filterwarnings("ignore", category=RuntimeWarning) 
    
    hp.orthview(
            map=data_map,coord='E', fig=fig,
            half_sky=True, rot=(0,90,180), xsize=600,
            title=title,sub=sub,min=vmin,max=vmax,
            cmap=cmap,notext=True, hold=True, cbar=cbar)

    hp.graticule(dpar=10,coord='E',color='k',alpha=0.3,dmer=45)
   
    # Altitude grid
    hp.projtext(00.0*(np.pi/180.0), 225.0*(np.pi/180),'0', color='w', coord='E')
    hp.projtext(30.0*(np.pi/180.0), 225.0*(np.pi/180),'30', color='w', coord='E')
    hp.projtext(60.0*(np.pi/180.0), 225.0*(np.pi/180),'60', color='w', coord='E')

    # NSEW 
    hp.projtext(80.0*(np.pi/180.0), 000.0*(np.pi/180.0), r'$N  $', coord='E',color='w', verticalalignment='top')
    hp.projtext(80.0*(np.pi/180.0), 090.0*(np.pi/180.0), r'$E  $', coord='E',color='w', horizontalalignment='right')
    hp.projtext(80.0*(np.pi/180.0), 180.0*(np.pi/180.0), r'$S  $', coord='E',color='w', verticalalignment='bottom')
    hp.projtext(80.0*(np.pi/180.0), 270.0*(np.pi/180.0), r'$W  $', coord='E',color='w', horizontalalignment='left')


if __name__ == "__main__":

    # Healpix constants
    nside = 32
    len_empty_healpix = hp.nside2npix(nside)   #12288
    
    # Had problems with having coords = 0 when interpolating, so make a small number
    # and call it epsilon for some reason
    
    epsilon = 1.e-12
    
    # Reference FEE models in XX & YY pols
    refXX = '../../data/FEE_Reference_Models/MWA_reference_tile_FarField_XX.ffe'
    refYY = '../../data/FEE_Reference_Models/MWA_reference_tile_FarField_YY.ffe'
    
    # Output directory
    out_dir = '../../outputs/reproject_ref'
    Path(out_dir).mkdir(parents=True, exist_ok=True)
    
    healpix_XX,theta_mesh,phi_mesh,power_XX,theta = create_model(len_empty_healpix, epsilon, nside, file_name=refXX)
    healpix_YY,theta_mesh,phi_mesh,power_YY,theta = create_model(len_empty_healpix, epsilon, nside, file_name=refYY)
    
    # Plot the things to sanity check and save results
    fig = plt.figure(figsize=(10,10))
    
    ax1 = fig.add_subplot(221,projection='polar')
    ax1.set_rgrids([10,30,50,70,90], angle=22)
    ax2 = fig.add_subplot(222,projection='polar')
    ax2.set_rgrids([10,30,50,70,90], angle=22)
    
    im1 = ax1.pcolormesh(phi_mesh*(np.pi/180.0),
            theta_mesh,power_XX,label='XX',
            vmin=-40,vmax=-20, cmap=cmap)
    
    ax1.set_title('Ref XX', y=1.1)
    ax1.grid(color='k',alpha=0.3)
    
    
    im2 = ax2.pcolormesh(phi_mesh*(np.pi/180.0),
            theta_mesh,power_YY,label='YY',
            vmin=-40,vmax=-20, cmap=cmap)
    
    ax2.set_title('Ref YY', y=1.1)
    ax2.grid(color='k',alpha=0.3)
    
    
    ax3 = fig.add_subplot(2,2,3)
    plot_healpix(data_map=healpix_XX,sub=(2,2,3),title=None,vmin=-20,vmax=0, cmap=cmap)
    ax3.axis('off')
    ax3.set_title('Healpix XX',x=0.4, y=-0.33)
    
    ax4 = fig.add_subplot(2,2,4)
    plot_healpix(data_map=healpix_YY,sub=(2,2,4),title=None,vmin=-20,vmax=0, cmap=cmap)
    ax4.axis('off')
    ax4.set_title('Healpix YY',x=0.6, y=-0.33)
    
    np.savez_compressed(f'{out_dir}/ref_dipole_models.npz',XX=healpix_XX, YY=healpix_YY)
    fig.savefig(f'{out_dir}/reproject_dipole_models.png',bbox_inches='tight')



# Package imports
from setuptools import setup, find_namespace_packages
from pathlib import Path
import re

# Get long description from README
with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()

# Get the requirements list
with open('requirements.txt', 'r') as f:
    requirements = f.read().splitlines()

# Read the __version__.py file
with open('embers/__version__.py', 'r') as f:
    vf = f.read()

# Recursive data files in embers.kindle
kindle_data = Path(f'{__file__}/embers/kindle/data')
files = [str(p.relative_to(kindle_data)) for p in kindle_data.rglob('*.txt')]

# Obtain version from read-in __version__.py file
version = re.search(r"^_*version_* = ['\"]([^'\"]*)['\"]", vf, re.M).group(1)


setup(
    name="embers",
    version="0.0.1",
    license='MIT',
    author="Aman Chokshi",
    author_email="achokshi@student.unimelb.edu.au",
    description="Experimental Measurement of BEam Response with Satellites",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/amanchokshi/EMBERS",
    packages=find_namespace_packages(include=["embers.*"]),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering :: Astronomy",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    keywords=("embers radio astronomy satellites beam measurement"),
    python_requires='>=3.6, <4',
    install_requires=requirements,
    include_package_data=True,
    package_data={
    "embers": ["data/*.ffe"],
    "embers.kindle": files,
    },
    zip_safe=False,
)



import os
import time
import argparse
from pathlib import Path
from sat_ids import norad_ids
import spacetrack.operators as op
from spacetrack import SpaceTrackClient

# space-tracks.org login and password, saved as environment variables

if os.environ.get('ST_USER') != None:
    st_user = os.environ.get('ST_USER')
    st_pass = os.environ.get('ST_PASS')

    st = SpaceTrackClient(identity=st_user, password=st_pass)
    
    parser = argparse.ArgumentParser(description="""Downloads TLE files for the Beam Experiment from Space-tracks.org.""")
    parser.add_argument('--start_date', metavar='\b', help='Start date in format YYYY-MM-DD')
    parser.add_argument('--stop_date', metavar='\b', help='Stop date in format YYYY-MM-DD')
    parser.add_argument('--out_dir', metavar='\b', default='./../../outputs/sat_ephemeris/TLE/', help='Path to output directory. Default=./../../outputs/sat_ephemeris/TLE/')

    args = parser.parse_args()

    start_date = args.start_date
    stop_date = args.stop_date
    out_dir = args.out_dir
    
    
    
    #make a TLE directory
    Path(out_dir).mkdir(parents=True, exist_ok=True)
    os.makedirs(os.path.dirname(out_dir), exist_ok=True)
    
    print('Starting TLE download')
    
    for sat_name, sat_id in norad_ids.items():
        #Sleep to limit downloads to 200 TLEs per hour
        time.sleep(20)
        data = st.tle(iter_lines=True, norad_cat_id=sat_id, orderby='epoch asc', epoch=f'{start_date}--{stop_date}', format='tle')
        print(f'downloading tle for {sat_name} satellite [{sat_id}] from space-tracks.org')
        with open(f'{out_dir}/{sat_id}.txt', 'w') as fp:
            for line in data:
                fp.write(line + '\n')
    
            
else:
    print('ERROR!')
    print('Space-Track.org credentials do not exist!')
    print('Edit this file and enter your account details.')



import sys
import numpy as np
import healpy as hp
import scipy.optimize as opt
import numpy.polynomial.polynomial as poly
from mwa_pb.mwa_sweet_spots import all_grid_points

sys.path.append('../../decode_rf_data')
from colormap import spectral, jade, kelp

# Custom spectral colormap
jade, _ = jade()

def hp_slices_horizon(nside=None, zenith_angle=90):
    '''Healpix pix indices of NS, EW slices and above horizon'''
    # theta phi values of each pixel 
    hp_indices = np.arange(hp.nside2npix(nside))
    θ, ɸ = hp.pix2ang(nside, hp_indices)

    # healpix indices above the horizon
    above_horizon_indices = np.where(θ <= np.radians(zenith_angle))[0]
    
    # pixel coords above the horizon
    ɸ_above_horizon = ɸ[above_horizon_indices]

    NS_indices = []
    EW_indices = []

    # pixel indices along N, E, S, W slices
    # order the indices such that they proceed from N -> S or E -> W
    n_slice = sorted(np.where((np.round(np.degrees(ɸ_above_horizon))) ==  45)[0])
    e_slice = sorted(np.where((np.round(np.degrees(ɸ_above_horizon))) == 135)[0])
    s_slice = sorted(np.where((np.round(np.degrees(ɸ_above_horizon))) == 225)[0], reverse=True)
    w_slice = sorted(np.where((np.round(np.degrees(ɸ_above_horizon))) == 315)[0], reverse=True)

    NS_indices.extend(s_slice)
    NS_indices.extend(n_slice)
    EW_indices.extend(w_slice)
    EW_indices.extend(e_slice)

    return [NS_indices, EW_indices, above_horizon_indices]

def slice_map(hp_map, za):
    '''slices healpix map along NS, EW'''
    
    NS_indices, EW_indices, _ = hp_slices_horizon(nside, zenith_angle=za)

    θ_NS, ɸ_NS = np.degrees(hp.pix2ang(nside, NS_indices))
    θ_EW, ɸ_EW = np.degrees(hp.pix2ang(nside, EW_indices))

    zenith_angle_NS = []
    for i, j in zip(θ_NS, ɸ_NS):
        if j <= 180:
            zenith_angle_NS.append(-1*i)
        else:
            zenith_angle_NS.append(i)
    
    zenith_angle_EW = []
    for i, j in zip(θ_EW, ɸ_EW):
        if j <= 180:
            zenith_angle_EW.append(-1*i)
        else:
            zenith_angle_EW.append(i)
    
    NS_data = [hp_map[NS_indices], zenith_angle_NS]
    EW_data = [hp_map[EW_indices], zenith_angle_EW]

    return [NS_data, EW_data]

def nan_mad(ref_map):
    '''Compute mad while ignoring nans'''
    ref_map_mad = []
    for j in ref_map:
        if j != []:
            j = np.asarray(j)
            j = j[~np.isnan(j)]
            ref_map_mad.append(mad(j))
        else:
            ref_map_mad.append(np.nan)

    ref_map_mad = np.asarray(ref_map_mad)
    ref_map_mad[np.where(ref_map_mad == np.nan)] = np.nanmean(ref_map_mad)

    return ref_map_mad


def tile_map_slice(good_map, za):
    '''slices ref healpix map along NS & EW'''
    
    ref_map_NS, ref_map_EW = slice_map(np.asarray(good_map), za)

    ref_med_map_NS = np.asarray([(np.nanmedian(i) if i != [] else np.nan) for i in ref_map_NS[0]])
    ref_mad_map_NS = np.asarray(nan_mad(ref_map_NS[0]))
    za_NS = ref_map_NS[1]

    ref_med_map_EW = np.asarray([(np.nanmedian(i) if i != [] else np.nan) for i in ref_map_EW[0]])
    ref_mad_map_EW = np.asarray(nan_mad(ref_map_EW[0]))
    za_EW = ref_map_EW[1]
    
    NS_data = [ref_med_map_NS, ref_mad_map_NS, za_NS]
    EW_data = [ref_med_map_EW, ref_mad_map_EW, za_EW]

    return [NS_data, EW_data]

# rotate func written by Jack Line
def rotate(nside, angle=None,healpix_array=None,savetag=None,flip=False):
    '''Takes in a healpix array, rotates it by the desired angle, and saves it.
    Optionally flip the data, changes east-west into west-east because
    astronomy'''

    # theta phi values of each pixel 
    hp_indices = np.arange(hp.nside2npix(nside))
    θ, ɸ = hp.pix2ang(nside, hp_indices)

    new_hp_inds = hp.ang2pix(nside,θ,ɸ+angle)

    ##Flip the data to match astro conventions
    if flip == True:
        new_angles = []
        for phi in ɸ:
            if phi <= np.pi:
                new_angles.append(np.pi - phi)
            else:
                new_angles.append(3*np.pi - phi)
        new_hp_inds = hp.ang2pix(nside,ɸ, np.asarray(new_angles))

    ##Save the array in the new order
    if savetag:
        np.savez_compressed(savetag,beammap=healpix_array[new_hp_inds])

    return healpix_array[new_hp_inds]

# chisquared minimization to best fit map to data
def fit_gain(map_data=None, map_error=None,beam=None):
    '''Fit the beam model to the measured data using
    chisquared minimization'''
    
    bad_values = np.isnan(map_data)
    map_data = map_data[~bad_values]
    map_error = map_error[~bad_values]

    map_error[np.where(map_error == 0)] = np.mean(map_error)

    def chisqfunc(gain):
        model = beam[~bad_values] + gain
        chisq = sum((map_data - model)**2)
        #chisq = sum(((map_data - model)/map_error)**2)
        return chisq

    x0 = np.array([0])

    result =  opt.minimize(chisqfunc,x0)
    
    return result.x

def poly_fit(x, y, map_data, order):
    '''Fit polynominal of order to data'''
    
    x = np.asarray(x)
    y = np.asarray(y)

    bad_values = np.isnan(map_data)
    x_good = x[~bad_values]
    y_good = y[~bad_values]
    coefs = poly.polyfit(x_good, y_good, order)
    fit = poly.polyval(x, coefs)
    return fit


def plt_slice(
        fig=None, sub=[1, 1, 1],
        zen_angle=None, map_slice=None,
        map_error=None, model_slice=None, 
        delta_pow=None, pow_fit=None, 
        slice_label=None, model_label=None,
        xlabel=False, ylabel=False, title=None):

    '''Plot a slice of the beam, with measured
    data, errorbars, and fit the simulated beam
    to the data. Also plot the diff b/w data and
    the model'''

    ax = fig.add_subplot(sub[0], sub[1], sub[2])

    ax.errorbar(
           zen_angle, map_slice, yerr=map_error, 
           fmt='.', color='#a8df65', ecolor='#7da87b',
           elinewidth=2.1, capsize=2.1, capthick=2.6,
           alpha=0.9, label=slice_label)

    ax.plot(zen_angle,model_slice, color='#c70039', linewidth=2.1, alpha=0.9, label=model_label)
    ax.set_xticks( [-75,-50,-25,0,25,50,75 ])

    leg = ax.legend(loc="lower center", frameon=True, framealpha=0.3, handlelength=1)
    leg.get_frame().set_facecolor('white')
    for l in leg.legendHandles:
        l.set_alpha(1)
    for text in leg.get_texts():
        plt.setp(text, color = 'w')

    #ax.set_ylim(bottom=-30)
    ax.set_xlim([-92,92])
    ax.set_ylim([-50,5])
    ax.set_title(title, loc='left', color='w')

    divider = make_axes_locatable(ax)
    dax = divider.append_axes("bottom", size="30%", pad=0.05)
    dax.set_xticks([-75,-50,-25,0,25,50,75 ])

    if xlabel is True:
        dax.set_xlabel('Zenith Angle [deg]')
    else:
        dax.set_xticklabels([])
        ax.set_xticklabels([])

    if ylabel is True:
        ax.set_ylabel('Power [dB]')
        ax.set_yticks([-40, -30, -20, -10, 0])
        dax.set_ylabel(r'$\Delta$P [dB]')
    else:
        dax.set_yticklabels([])
        ax.set_yticklabels([])

    #dax = fig.add_subplot(2,1,2)
    dax.scatter(zen_angle, delta_pow, marker='.', s=36, alpha=0.9, color='#63b7af')
    dax.plot(zen_angle, pow_fit, linewidth=2.1, alpha=0.9, color='#ff8264')
    dax.set_ylim([-10,10])
    #dax.set_xticklabels([])
    dax.set_xlim([-92,92])

    return ax

        
def beam_maps(f):
    '''Returns pairs of [NS,EW] slices of maps, for each pointing'''
    f_name, _ = Path(f).name.split('.')
    t_name, r_name, _, _ = f_name.split('_')

    pointings = ['0','2','4']
    
    maps = []

    # load data from map .npz file
    tile_map = np.load(f, allow_pickle=True)
    fee_m = np.load(fee_map, allow_pickle=True)
    
    for p in pointings:

        tile = tile_map[p]

        if 'XX' in t_name:
            fee = fee_m[p][0]
        else:
            fee = fee_m[p][1]
        

        # rotate maps so slices can be taken 
        fee_r = rotate(nside, angle= -np.pi/4, healpix_array=fee)
        tile_r = rotate(nside, angle=-np.pi/4, healpix_array=tile)

        # slice the tile and fee maps along NS, EW
        # zenith angle thresh of 70 to determine fit gain factor
        NS_f, EW_f = slice_map(fee_r, 70)
        NS_t, EW_t = tile_map_slice(tile_r, 70)

        gain_NS = fit_gain(map_data=NS_t[0], map_error=NS_t[1], beam=NS_f[0])
        gain_EW = fit_gain(map_data=EW_t[0], map_error=EW_t[1], beam=EW_f[0])

        # slice the tile and fee maps along NS, EW. 
        # the above gain factor is applied to full beam slices
        NS_fee, EW_fee = slice_map(fee_r, 90)
        NS_tile, EW_tile = tile_map_slice(tile_r, 90)

        # Scale the data so that it best fits the beam slice
        NS_tile_med = NS_tile[0] - gain_NS[0]
        EW_tile_med = EW_tile[0] - gain_EW[0]
        
        # delta powers
        del_NS = NS_tile_med - NS_fee[0]
        del_EW = EW_tile_med - EW_fee[0]

        # 3rd order poly fits for residuals
        fit_NS = poly_fit(NS_tile[2], del_NS, NS_tile[0], 3)  
        fit_EW = poly_fit(EW_tile[2], del_EW, EW_tile[0], 3)  

       
        maps.append([[NS_tile, NS_fee, NS_tile_med, del_NS, fit_NS],[EW_tile, EW_fee, EW_tile_med, del_EW, fit_EW]])
        
    return maps


def plt_grid(tile_name, ref_name):
    # This is an Awesome plot

    f_xx = f'{map_dir}/{tile_name}XX_{ref_name}XX_tile_maps.npz'
    f_yy = f'{map_dir}/{tile_name}YY_{ref_name}YY_tile_maps.npz'
    
    maps_xx = beam_maps(f_xx) 
    maps_yy = beam_maps(f_yy)
    
    
    nice_fonts = {
            # Use LaTeX to write all text
            #"text.usetex": True,
            "font.family": "sans-serif",
            # Use 10pt font in plots, to match 10pt font in document
            "axes.labelsize": 14,
            "axes.titlesize": 12,
            "font.size": 14,
            # Make the legend/label fonts a little smaller
            "legend.fontsize": 9,
            "xtick.labelsize": 10,
            "ytick.labelsize": 10,
            "ytick.color" : "w",
            "xtick.color" : "w",
            "axes.labelcolor" : "w",
            "axes.edgecolor" : "w"
            }
    
    plt.rcParams.update(nice_fonts)
    
    fig1 = plt.figure(figsize=(11,6))

    #ax1  = fig1.add_axes([0.01, 0.75, 0.29, 0.22])
    
    ax1 = plt.subplot(2,3,1)
    plt_slice(
            fig=fig1, sub=[2,3,1],
            zen_angle=      maps_xx[0][0][0][2],
            map_slice=      maps_xx[0][0][2],
            map_error=      maps_xx[0][0][0][1],
            model_slice=    maps_xx[0][0][1][0],
            delta_pow=      maps_xx[0][0][3],
            pow_fit=        maps_xx[0][0][4], 
            slice_label=    'Tile NS',
            model_label=    'FEE NS',
            ylabel=         True,
            title=          fr'($i$) NS slice of {tile_name}XX [zenith]')
    
    ax2 = plt.subplot(2,3,2)
    plt_slice(
            fig=fig1, sub=[2,3,2],
            zen_angle=      maps_xx[1][0][0][2],
            map_slice=      maps_xx[1][0][2],
            map_error=      maps_xx[1][0][0][1],
            model_slice=    maps_xx[1][0][1][0],
            delta_pow=      maps_xx[1][0][3],
            pow_fit=        maps_xx[1][0][4], 
            slice_label=    'Tile NS',
            model_label=    'FEE NS',
            title=          fr'($ii$) NS slice of {tile_name}XX [2]')
    
    ax3 = plt.subplot(2,3,3)
    plt_slice(
            fig=fig1, sub=[2,3,3],
            zen_angle=      maps_xx[2][0][0][2],
            map_slice=      maps_xx[2][0][2],
            map_error=      maps_xx[2][0][0][1],
            model_slice=    maps_xx[2][0][1][0],
            delta_pow=      maps_xx[2][0][3],
            pow_fit=        maps_xx[2][0][4], 
            slice_label=    'Tile NS',
            model_label=    'FEE NS',
            title=          fr'($iii$) NS slice of {tile_name}XX [4]')
    
    ax4 = plt.subplot(2,3,4)
    plt_slice(
            fig=fig1, sub=[2,3,4],
            zen_angle=      maps_xx[0][1][0][2],
            map_slice=      maps_xx[0][1][2],
            map_error=      maps_xx[0][1][0][1],
            model_slice=    maps_xx[0][1][1][0],
            delta_pow=      maps_xx[0][1][3],
            pow_fit=        maps_xx[0][1][4], 
            slice_label=    'Tile EW',
            model_label=    'FEE EW',
            ylabel=         True,
            xlabel=         True,
            title=          fr'($iv$) EW slice of {tile_name}XX [zenith]')
    
    ax5 = plt.subplot(2,3,5)
    plt_slice(
            fig=fig1, sub=[2,3,5],
            zen_angle=      maps_xx[1][1][0][2],
            map_slice=      maps_xx[1][1][2],
            map_error=      maps_xx[1][1][0][1],
            model_slice=    maps_xx[1][1][1][0],
            delta_pow=      maps_xx[1][1][3],
            pow_fit=        maps_xx[1][1][4], 
            slice_label=    'Tile EW',
            model_label=    'FEE EW',
            xlabel=         True,
            title=          fr'($v$) EW slice of {tile_name}XX [2]')
    
    ax6 = plt.subplot(2,3,6)
    plt_slice(
            fig=fig1, sub=[2,3,6],
            zen_angle=      maps_xx[2][1][0][2],
            map_slice=      maps_xx[2][1][2],
            map_error=      maps_xx[2][1][0][1],
            model_slice=    maps_xx[2][1][1][0],
            delta_pow=      maps_xx[2][1][3],
            pow_fit=        maps_xx[2][1][4], 
            slice_label=    'Tile EW',
            model_label=    'FEE EW',
            xlabel=         True,
            title=          fr'($vi$) EW slice of {tile_name}XX [4]')
    
    
    plt.tight_layout() 
    plt.savefig(f'{tile_name}_{ref_name}_map_slices.png', transparent=True, bbox_inches='tight', dpi=420)
    plt.close()


if __name__=='__main__':
    
    import argparse
    import numpy as np
    from pathlib import Path
    import concurrent.futures
    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
    import matplotlib.gridspec as gs
    from mpl_toolkits.axes_grid1 import make_axes_locatable
    from scipy.stats import median_absolute_deviation as mad

    import sys
    sys.path.append('../decode_rf_data')
    from colormap import spectral
    
    # Custom spectral colormap
    cmap = spectral()
    
    parser = argparse.ArgumentParser(description="""
        Plot healpix map of reference data
        """)
    
    parser.add_argument('--out_dir', metavar='\b', default='../../../outputs/paper_plots/tile_grids/',
            help='Output directory. Default=../../../outputs/paper_plots/tile_grids')
    parser.add_argument('--map_dir', metavar='\b', default='../../../outputs/tile_maps/tile_maps_norm/',
            help='Tile map directory. Default=../../../outputs/tile_maps/tile_maps_norm/')
    parser.add_argument('--fee_map', metavar='\b', default='../../../outputs/tile_maps/FEE_maps/mwa_fee_beam.npz',
            help='Healpix FEE map of mwa tile. default=../../../outputs/tile_maps/FEE_maps/mwa_fee_beam.npz')
    parser.add_argument('--nside', metavar='\b',type=int, default=32,help='Healpix Nside. Default = 32')
    
    
    args = parser.parse_args()
    
    out_dir     = Path(args.out_dir)
    map_dir     = Path(args.map_dir)
    fee_map     = args.fee_map
    nside       = args.nside
    
    # make output dir if it doesn't exist
    out_dir.mkdir(parents=True, exist_ok=True)

    tiles = [
            'S06', 'S07', 'S08', 'S09',
            'S10', 'S12', 'S29', 'S30',
            'S31', 'S32', 'S33', 'S34',
            'S35', 'S36'
            ]

    refs = ['rf0', 'rf1']
    
    #for r in refs:
    #    for t in tiles:
    #        plt_grid(t, r)

    plt_grid('S07','rf1')





import numpy as np
import healpy as hp
import scipy.optimize as opt
import numpy.polynomial.polynomial as poly


def hp_slices_horizon(nside=None):
    '''Healpix pix indices of NS, EW slices and above horizon'''
    # theta phi values of each pixel 
    hp_indices = np.arange(hp.nside2npix(nside))
    θ, ɸ = hp.pix2ang(nside, hp_indices)

    # healpix indices above the horizon
    above_horizon_indices = np.where(θ <= np.radians(80))[0]
    
    # pixel coords above the horizon
    ɸ_above_horizon = ɸ[above_horizon_indices]

    NS_indices = []
    EW_indices = []

    # pixel indices along N, E, S, W slices
    # order the indices such that they proceed from N -> S or E -> W
    n_slice = sorted(np.where((np.round(np.degrees(ɸ_above_horizon))) ==  45)[0], reverse=True)
    e_slice = sorted(np.where((np.round(np.degrees(ɸ_above_horizon))) == 135)[0], reverse=True)
    s_slice = sorted(np.where((np.round(np.degrees(ɸ_above_horizon))) == 225)[0])
    w_slice = sorted(np.where((np.round(np.degrees(ɸ_above_horizon))) == 315)[0])

    NS_indices.extend(n_slice)
    NS_indices.extend(s_slice)
    EW_indices.extend(e_slice)
    EW_indices.extend(w_slice)

    return [NS_indices, EW_indices, above_horizon_indices]


def slice_map(hp_map):
    '''slices healpix map along NS, EW'''
    
    NS_indices, EW_indices, _ = hp_slices_horizon(nside)

    θ_NS, ɸ_NS = np.degrees(hp.pix2ang(nside, NS_indices))
    θ_EW, ɸ_EW = np.degrees(hp.pix2ang(nside, EW_indices))

    zenith_angle_NS = []
    for i, j in zip(θ_NS, ɸ_NS):
        if j <= 180:
            zenith_angle_NS.append(-1*i)
        else:
            zenith_angle_NS.append(i)
    
    zenith_angle_EW = []
    for i, j in zip(θ_EW, ɸ_EW):
        if j <= 180:
            zenith_angle_EW.append(-1*i)
        else:
            zenith_angle_EW.append(i)
    
    NS_data = [hp_map[NS_indices], zenith_angle_NS]
    EW_data = [hp_map[EW_indices], zenith_angle_EW]

    return [NS_data, EW_data]

def nan_mad(ref_map):
    '''Compute mad while ignoring nans'''
    ref_map_mad = []
    for j in ref_map:
        if j != []:
            j = np.asarray(j)
            j = j[~np.isnan(j)]
            ref_map_mad.append(mad(j))
        else:
            ref_map_mad.append(np.nan)

    ref_map_mad = np.asarray(ref_map_mad)
    ref_map_mad[np.where(ref_map_mad == np.nan)] = np.nanmean(ref_map_mad)

    return ref_map_mad


def good_maps(ref_map):
    '''Creates a ref map with only good satellites'''
   
    pointings = ['0','2','4']
    
    # load data from map .npz file
    f = Path(f'{map_dir}/{ref_map}')
    tile_data = np.load(f, allow_pickle=True)
    tile_data = {key:tile_data[key].item() for key in tile_data}
    ref_map = tile_data['ref_map']  
    
    # Good sats from which to make plots
    good_sats = [
            25338, 25984, 25985,
            28654, 40086, 40087,
            40091, 41179, 41180,
            41182, 41183, 41184,
            41185, 41187, 41188,
            41189, 44387
            ]
    
    # Empty good map
    good_map = [[] for pixel in range(hp.nside2npix(nside))]
    
    for p in pointings:
        
        # append to good map from all good sat data
        for sat in good_sats:
            for pix in range(hp.nside2npix(nside)):
                good_map[pix].extend(ref_map[p][sat][pix])
        
    return good_map

def ref_map_slice(good_map):
    '''slices ref healpix map along NS & EW'''
    
    ref_map_NS, ref_map_EW = slice_map(np.asarray(good_map))

    ref_med_map_NS = np.asarray([(np.nanmedian(i) if i != [] else np.nan) for i in ref_map_NS[0]])
    # Scale mean map such that the max value is 0
    ref_med_map_scaled_NS = np.asarray([i-np.nanmax(ref_med_map_NS) for i in ref_med_map_NS])
    #ref_mad_map_NS = np.asarray([mad(i) for i in ref_map_NS[0]])
    ref_mad_map_NS = np.asarray(nan_mad(ref_map_NS[0]))
    za_NS = ref_map_NS[1]

    ref_med_map_EW = np.asarray([(np.nanmedian(i) if i != [] else np.nan) for i in ref_map_EW[0]])
    # Scale mean map such that the max value is 0
    ref_med_map_scaled_EW = np.asarray([i-np.nanmax(ref_med_map_EW) for i in ref_med_map_EW])
    #ref_mad_map_EW = np.asarray([mad(i) for i in ref_map_EW[0]])
    ref_mad_map_EW = np.asarray(nan_mad(ref_map_EW[0]))
    za_EW = ref_map_EW[1]
    
    NS_data = [ref_med_map_scaled_NS, ref_mad_map_NS, za_NS]
    EW_data = [ref_med_map_scaled_EW, ref_mad_map_EW, za_EW]

    return [NS_data, EW_data]


# rotate func written by Jack Line
def rotate(nside, angle=None,healpix_array=None,savetag=None,flip=False):
    '''Takes in a healpix array, rotates it by the desired angle, and saves it.
    Optionally flip the data, changes east-west into west-east because
    astronomy'''

    # theta phi values of each pixel 
    hp_indices = np.arange(hp.nside2npix(nside))
    θ, ɸ = hp.pix2ang(nside, hp_indices)

    new_hp_inds = hp.ang2pix(nside,θ,ɸ+angle)

    ##Flip the data to match astro conventions
    if flip == True:
        new_angles = []
        for phi in ɸ:
            if phi <= np.pi:
                new_angles.append(np.pi - phi)
            else:
                new_angles.append(3*np.pi - phi)
        new_hp_inds = hp.ang2pix(nside,ɸ, np.asarray(new_angles))

    ##Save the array in the new order
    if savetag:
        np.savez_compressed(savetag,beammap=healpix_array[new_hp_inds])

    return healpix_array[new_hp_inds]

# chisquared minimization to best fit map to data
def fit_gain(map_data=None,map_error=None,beam=None):
    '''Fit the beam model to the measured data using
    chisquared minimization'''

    bad_values = np.isnan(map_data)
    map_data = map_data[~bad_values]
    map_error = map_error[~bad_values]

    map_error[np.where(map_error == 0)] = np.mean(map_error)

    def chisqfunc(gain):
        model = beam[~bad_values] + gain
        chisq = sum((map_data - model)**2)
        #chisq = sum(((map_data - model)/map_error)**2)
        return chisq

    x0 = np.array([0])

    result =  opt.minimize(chisqfunc,x0)
    
    return result.x


def poly_fit(x, y, map_data, order):
    '''Fit polynominal of order to data'''
    
    x = np.asarray(x)
    y = np.asarray(y)

    bad_values = np.isnan(map_data)
    x_good = x[~bad_values]
    y_good = y[~bad_values]
    coefs = poly.polyfit(x_good, y_good, order)
    fit = poly.polyval(x, coefs)
    return fit


def plt_slice(
        fig=None, sub=(None, None, None),
        zen_angle=None, map_slice=None,
        map_error=None, model_slice=None, 
        delta_pow=None, pow_fit=None, 
        slice_label=None, model_label=None,
        ylabel=True, title=None):

    '''Plot a slice of the beam, with measured
    data, errorbars, and fit the simulated beam
    to the data. Also plot the diff b/w data and
    the model'''

    ax = fig.add_subplot(sub[0], sub[1], sub[2])

    ax.errorbar(
           zen_angle, map_slice, yerr=map_error, 
           fmt='.', color='#326765', ecolor='#7da87b',
           elinewidth=1.4, capsize=1.4, capthick=1.6,
           alpha=0.9, ms=7, label=slice_label)

    ax.plot(zen_angle,model_slice, color='#c70039', linewidth=1.4, alpha=0.9, label=model_label)

    ax.text(0.02,.88,title,
            horizontalalignment='left',
            transform=ax.transAxes)

    #ax.set_ylim(bottom=-30)
    #ax.set_xlabel('Zenith Angle (degrees)')
    #ax.legend(loc='lower center')
    leg = ax.legend(loc="lower center", frameon=True, handlelength=1)
    leg.get_frame().set_facecolor('white')
    for l in leg.legendHandles:
        l.set_alpha(1)
    ax.set_xlim([-82,82])
    ax.set_ylim([-26,12])

    divider = make_axes_locatable(ax)
    dax = divider.append_axes("bottom", size="30%", pad=0.1)

    #dax = fig.add_subplot(2,1,2)
    dax.scatter(zen_angle, delta_pow, marker='.', s=30, color='#27296d')
    dax.plot(zen_angle, pow_fit, linewidth=1.4, alpha=0.9, color='#ff8264')
    dax.set_xlim([-82,82])
    dax.set_xticklabels([])
    dax.set_ylim([-5,5])
    if ylabel is True:
        ax.set_ylabel('Power [dB]')
        dax.set_ylabel(r'$\Delta$ref [dB]')
        #dax.set_ylabel('Data - Model (dB)')
    else:
        dax.set_yticklabels([])
        ax.set_yticklabels([])

    return ax

def plt_null_test(
        fig=None, sub=(None, None, None), 
        zen_angle=None, del_pow=None, 
        del_err=None, del_beam=None, 
        del_fit=None, null_label=None,
        beam_label=None, fit_label=None,
        ylabel=True, title=None):

    '''Plot graphs for a null test'''

    ax = fig.add_subplot(sub[0], sub[1], sub[2])
    
    ax.errorbar(
           zen_angle, del_pow, yerr=del_err, 
           fmt='.', color='#326765', ecolor='#7da87b',
           elinewidth=1.4, capsize=1.4, capthick=1.6,
           alpha=0.9, ms=7, label=null_label)
    
    ax.text(0.02,0.9,title,
            horizontalalignment='left',
            transform=ax.transAxes)

    ax.plot(zen_angle,del_beam, color='#c70039', linewidth=1.4, alpha=0.9, label=beam_label)
    
    ax.plot(zen_angle,del_fit, color='#ff8264', linewidth=1.4, alpha=0.9, label=fit_label)
    
    ax.set_xlim([-82,82])
    ax.set_ylim([-10,10])
    ax.set_xlabel('Zenith Angle [degrees]')
    #ax.legend(loc='upper left')
    leg = ax.legend(loc="lower center", frameon=True, handlelength=1)
    leg.get_frame().set_facecolor('white')
    for l in leg.legendHandles:
        l.set_alpha(1)
    
    if ylabel is True:
        ax.set_ylabel('Power [dB]')
    else:
        ax.set_yticklabels([])

    return ax


if __name__=='__main__':
    
    import argparse
    import numpy as np
    from pathlib import Path
    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
    import matplotlib.gridspec as gs
    from mpl_toolkits.axes_grid1 import make_axes_locatable
    from scipy.stats import median_absolute_deviation as mad

    import sys
    sys.path.append('../decode_rf_data')
    sys.path.append('../tile_maps')
    from plot_tile_maps import plot_healpix
    from colormap import spectral
    
    # Custom spectral colormap
    cmap = spectral()
    
    parser = argparse.ArgumentParser(description="""
        Plot healpix map of reference data
        """)
    
    parser.add_argument('--out_dir', metavar='\b', default='../../outputs/tile_maps/null_test/',
            help='Output directory. Default=../../outputs/tile_maps/null_test/')
    parser.add_argument('--map_dir', metavar='\b', default='../../outputs/tile_maps/tile_maps_raw/',
            help='Output directory. Default=../../outputs/tile_maps/tile_maps_raw/')
    parser.add_argument('--ref_model', metavar='\b', default='../../outputs/reproject_ref/ref_dipole_models.npz',
            help='Healpix reference FEE model file. default=../../outputs/reproject_ref/ref_dipole_models.npz')
    parser.add_argument('--nside', metavar='\b',type=int, default=32,help='Healpix Nside. Default = 32')
    
    
    args = parser.parse_args()
    
    out_dir     = Path(args.out_dir)
    map_dir     = Path(args.map_dir)
    ref_model   = args.ref_model
    nside       = args.nside
    
    out_dir.mkdir(parents=True, exist_ok=True)
    
    ref_tiles = [
            'S35XX_rf0XX_sat_maps.npz', 
            'S35YY_rf0YY_sat_maps.npz', 
            'S35XX_rf1XX_sat_maps.npz', 
            'S35YY_rf1YY_sat_maps.npz'
            ]
    good_rf0XX = rotate(nside, angle=+(1*np.pi)/4.0, healpix_array=np.asarray(good_maps(ref_tiles[0])))
    good_rf0YY = rotate(nside, angle=+(1*np.pi)/4.0, healpix_array=np.asarray(good_maps(ref_tiles[1])))
    good_rf1XX = rotate(nside, angle=+(1*np.pi)/4.0, healpix_array=np.asarray(good_maps(ref_tiles[2])))
    good_rf1YY = rotate(nside, angle=+(1*np.pi)/4.0, healpix_array=np.asarray(good_maps(ref_tiles[3])))

    # NS, EW slices of all four reference tiles 
    rf0XX_NS, rf0XX_EW = ref_map_slice(good_rf0XX)
    rf0YY_NS, rf0YY_EW = ref_map_slice(good_rf0YY)
    rf1XX_NS, rf1XX_EW = ref_map_slice(good_rf1XX)
    rf1YY_NS, rf1YY_EW = ref_map_slice(good_rf1YY)

    # Null test diff in power b/w rf0 & rf1
    ref01_XX_NS = rf0XX_NS[0] - rf1XX_NS[0]
    ref01_XX_EW = rf0XX_EW[0] - rf1XX_EW[0]
    ref01_YY_NS = rf0YY_NS[0] - rf1YY_NS[0]
    ref01_YY_EW = rf0YY_EW[0] - rf1YY_EW[0]
    
    # Error propogation in null test
    error_ref01_XX_NS = np.sqrt((rf0XX_NS[1])**2 + (rf1XX_NS[1])**2 )
    error_ref01_XX_EW = np.sqrt((rf0XX_EW[1])**2 + (rf1XX_EW[1])**2 )
    error_ref01_YY_NS = np.sqrt((rf0YY_NS[1])**2 + (rf1YY_NS[1])**2 )
    error_ref01_YY_EW = np.sqrt((rf0YY_EW[1])**2 + (rf1YY_EW[1])**2 )
    
    # Load reference FEE model
    ref_fee_model = np.load(ref_model, allow_pickle=True)
    beam_XX = ref_fee_model['XX']
    beam_YY = ref_fee_model['YY']
   
    # Rotate beam models by pi/4 to match rotation of data
    rotated_XX = rotate(nside, angle=-(1*np.pi)/4.0,healpix_array=beam_XX)
    rotated_YY = rotate(nside, angle=-(1*np.pi)/4.0,healpix_array=beam_YY)
    
    # These plots show that the pi/4 rotation was correct
    #plot_healpix(data_map=rotated_XX,sub=(1,1,1), cmap=cmap, vmin=-40, vmax=-20)
    #plot_healpix(data_map=rotated_YY,sub=(1,1,1), cmap=cmap, vmin=-40, vmax=-20)
    #plt.show()
    
    # slice the XX rotated map along NS, EW
    XX_NS, XX_EW = slice_map(rotated_XX)
    XX_NS_slice, za_NS = XX_NS
    XX_EW_slice, za_EW = XX_EW
    

    # slice the YY rotated map along NS, EW
    YY_NS, YY_EW = slice_map(rotated_YY)
    YY_NS_slice, za_NS = YY_NS
    YY_EW_slice, za_EW = YY_EW

    # Gain offsets for the 8 combinations of data and beam slices
    gain_ref0_XX_NS = fit_gain(map_data=rf0XX_NS[0], map_error=rf0XX_NS[1], beam=XX_NS_slice)
    gain_ref0_XX_EW = fit_gain(map_data=rf0XX_EW[0], map_error=rf0XX_EW[1], beam=XX_EW_slice)
    gain_ref1_XX_NS = fit_gain(map_data=rf1XX_NS[0], map_error=rf1XX_NS[1], beam=XX_NS_slice)
    gain_ref1_XX_EW = fit_gain(map_data=rf1XX_EW[0], map_error=rf1XX_EW[1], beam=XX_EW_slice)
    
    gain_ref0_YY_NS = fit_gain(map_data=rf0YY_NS[0], map_error=rf0YY_NS[1], beam=YY_NS_slice)
    gain_ref0_YY_EW = fit_gain(map_data=rf0YY_EW[0], map_error=rf0YY_EW[1], beam=YY_EW_slice)
    gain_ref1_YY_NS = fit_gain(map_data=rf1YY_NS[0], map_error=rf1YY_NS[1], beam=YY_NS_slice)
    gain_ref1_YY_EW = fit_gain(map_data=rf1YY_EW[0], map_error=rf1YY_EW[1], beam=YY_EW_slice)
    
   
    # Scale the data so that it best fits the beam slice
    rf0XX_NS = rf0XX_NS - gain_ref0_XX_NS
    rf0XX_EW = rf0XX_EW - gain_ref0_XX_EW
    rf0YY_NS = rf0YY_NS - gain_ref0_YY_NS
    rf0YY_EW = rf0YY_EW - gain_ref0_YY_EW
    rf1XX_NS = rf1XX_NS - gain_ref1_XX_NS
    rf1XX_EW = rf1XX_EW - gain_ref1_XX_EW
    rf1YY_NS = rf1YY_NS - gain_ref1_YY_NS
    rf1YY_EW = rf1YY_EW - gain_ref1_YY_EW
    
    # Difference b/w beam model slices. 
    # Always 0 because we have only one model for both rf0, rf1 
    beam_ref01_XX_NS = XX_NS_slice - XX_NS_slice
    beam_ref01_XX_EW = XX_EW_slice - XX_EW_slice
    beam_ref01_YY_NS = YY_NS_slice - YY_NS_slice
    beam_ref01_YY_EW = YY_EW_slice - YY_EW_slice
   
    # delta powers
    del_pow_ref0_XX_NS = rf0XX_NS[0] - XX_NS_slice 
    del_pow_ref0_XX_EW = rf0XX_EW[0] - XX_EW_slice
    del_pow_ref1_XX_NS = rf1XX_NS[0] - XX_NS_slice
    del_pow_ref1_XX_EW = rf1XX_EW[0] - XX_EW_slice
   
    del_pow_ref0_YY_NS = rf0YY_NS[0] - YY_NS_slice
    del_pow_ref0_YY_EW = rf0YY_EW[0] - YY_EW_slice
    del_pow_ref1_YY_NS = rf1YY_NS[0] - YY_NS_slice
    del_pow_ref1_YY_EW = rf1YY_EW[0] - YY_EW_slice


    # 3rd order poly fits for residuals
    fit_ref0_XX_NS = poly_fit(za_NS, del_pow_ref0_XX_NS,rf0XX_NS[0], 3)  
    fit_ref0_XX_EW = poly_fit(za_EW, del_pow_ref0_XX_EW,rf0XX_EW[0], 3) 
    fit_ref1_XX_NS = poly_fit(za_NS, del_pow_ref1_XX_NS,rf1XX_NS[0], 3) 
    fit_ref1_XX_EW = poly_fit(za_EW, del_pow_ref1_XX_EW,rf1XX_EW[0], 3) 
    
    fit_ref0_YY_NS = poly_fit(za_NS, del_pow_ref0_YY_NS,rf0YY_NS[0], 3) 
    fit_ref0_YY_EW = poly_fit(za_EW, del_pow_ref0_YY_EW,rf0YY_EW[0], 3) 
    fit_ref1_YY_NS = poly_fit(za_NS, del_pow_ref1_YY_NS,rf1YY_NS[0], 3) 
    fit_ref1_YY_EW = poly_fit(za_EW, del_pow_ref1_YY_EW,rf1YY_EW[0], 3) 


    # Difference of power b/w ref0 & ref1 fits
    fit_ref01_XX_NS = fit_ref0_XX_NS - fit_ref1_XX_NS
    fit_ref01_XX_EW = fit_ref0_XX_EW - fit_ref1_XX_EW
    fit_ref01_YY_NS = fit_ref0_YY_NS - fit_ref1_YY_NS
    fit_ref01_YY_EW = fit_ref0_YY_EW - fit_ref1_YY_EW



    plt.style.use('seaborn')
    nice_fonts = {
            # Use LaTeX to write all text
            #"text.usetex": True,
            "font.family": "sans-serif",
            # Use 10pt font in plots, to match 10pt font in document
            "axes.labelsize": 8,
            "font.size": 10,
            # Make the legend/label fonts a little smaller
            "legend.fontsize": 6,
            "xtick.labelsize": 8,
            "ytick.labelsize": 8,
            }
    
    plt.rcParams.update(nice_fonts)
    
    fig1 = plt.figure(figsize=(12,7)) 
    
    ax1 = plt_slice(
            fig=fig1, sub=(3,4,1),
            zen_angle=za_NS, map_slice=rf0XX_NS[0],
            map_error=rf0XX_NS[1], model_slice=XX_NS_slice,
            delta_pow=del_pow_ref0_XX_NS, pow_fit=fit_ref0_XX_NS, 
            slice_label='ref0XX NS', model_label='FEE XX NS',
            title=r'($i$)')

    ax2 = plt_slice(
            fig=fig1, sub=(3,4,2),
            zen_angle=za_EW, map_slice=rf0XX_EW[0],
            map_error=rf0XX_EW[1], model_slice=XX_EW_slice,
            delta_pow=del_pow_ref0_XX_EW, pow_fit=fit_ref0_XX_EW,
            slice_label='ref0XX EW', model_label='FEE XX EW', ylabel=False,
            title=r'($ii$)')

    ax3 = plt_slice(
            fig=fig1, sub=(3,4,5),
            zen_angle=za_NS, map_slice=rf1XX_NS[0],
            map_error=rf1XX_NS[1], model_slice=XX_NS_slice,
            delta_pow=del_pow_ref1_XX_NS, pow_fit=fit_ref1_XX_NS,
            slice_label='ref1XX NS', model_label='FEE XX NS',
            title=r'($v$)')

    ax4 = plt_slice(
            fig=fig1, sub=(3,4,6),
            zen_angle=za_EW, map_slice=rf1XX_EW[0],
            map_error=rf1XX_EW[1], model_slice=XX_EW_slice,
            delta_pow=del_pow_ref1_XX_EW, pow_fit=fit_ref1_XX_EW,
            slice_label='ref1XX EW', model_label='FEE XX EW', ylabel=False,
            title=r'($vi$)')

    ax5 = plt_null_test(
            fig=fig1,sub=(3,4,9), 
            zen_angle=za_NS, del_pow=ref01_XX_NS, 
            del_err=error_ref01_XX_NS, del_beam=beam_ref01_XX_NS, 
            del_fit=fit_ref01_XX_NS, null_label='NS rf0-rf1', 
            beam_label='FEE Null', fit_label='Fit rf0-rf1',
            title=r'($ix$)')
    
    ax6 = plt_null_test(
            fig=fig1,sub=(3,4,10), 
            zen_angle=za_EW, del_pow=ref01_XX_EW, 
            del_err=error_ref01_XX_EW, del_beam=beam_ref01_XX_EW, 
            del_fit=fit_ref01_XX_EW, null_label='EW rf0-rf1', 
            beam_label='FEE Null', fit_label='Fit rf0-rf1', ylabel=False,
            title=r'($x$)')
    
    ax7 = plt_slice(
            fig=fig1, sub=(3,4,3),
            zen_angle=za_NS, map_slice=rf0YY_NS[0],
            map_error=rf0YY_NS[1], model_slice=YY_NS_slice,
            delta_pow=del_pow_ref0_YY_NS, pow_fit=fit_ref0_YY_NS,
            slice_label='ref0YY NS', model_label='FEE YY NS', ylabel=False,
            title=r'($iii$)')
    
    ax8 = plt_slice(
            fig=fig1, sub=(3,4,4),
            zen_angle=za_EW, map_slice=rf0YY_EW[0],
            map_error=rf0YY_EW[1], model_slice=YY_EW_slice,
            delta_pow=del_pow_ref0_YY_EW, pow_fit=fit_ref0_YY_EW,
            slice_label='ref0YY EW', model_label='FEE YY EW', ylabel=False,
            title=r'($iv$)')
    
    ax9 = plt_slice(
            fig=fig1, sub=(3,4,7),
            zen_angle=za_NS, map_slice=rf1YY_NS[0],
            map_error=rf1YY_NS[1], model_slice=YY_NS_slice,
            delta_pow=del_pow_ref1_YY_NS, pow_fit=fit_ref1_YY_NS,
            slice_label='ref1YY NS', model_label='FEE YY NS', ylabel=False,
            title=r'($vii$)')
    
    ax10 = plt_slice(
            fig=fig1, sub=(3,4,8),
            zen_angle=za_EW, map_slice=rf1YY_EW[0],
            map_error=rf1YY_EW[1], model_slice=YY_EW_slice,
            delta_pow=del_pow_ref1_YY_EW, pow_fit=fit_ref1_YY_EW,
            slice_label='ref1YY EW', model_label='FEE YY EW', ylabel=False,
            title=r'($viii$)')
    
    ax11 = plt_null_test(
            fig=fig1,sub=(3,4,11), 
            zen_angle=za_NS, del_pow=ref01_YY_NS, 
            del_err=error_ref01_YY_NS, del_beam=beam_ref01_YY_NS, 
            del_fit=fit_ref01_YY_NS, null_label='NS rf0-rf1', 
            beam_label='FEE Null', fit_label='Fit rf0-rf1', ylabel=False,
            title=r'($xi$)')
    
    ax12 = plt_null_test(
            fig=fig1,sub=(3,4,12), 
            zen_angle=za_EW, del_pow=ref01_YY_EW, 
            del_err=error_ref01_YY_EW, del_beam=beam_ref01_YY_EW, 
            del_fit=fit_ref01_YY_EW, null_label='EW rf0-rf1', 
            beam_label='FEE Null', fit_label='Fit rf0-rf1', ylabel=False,
            title=r'($xii$)')

    plt.tight_layout()
    fig1.savefig(f'../../outputs/paper_plots/null_test.pdf', bbox_inches='tight')
    
    



import sys
import json
import argparse
import matplotlib
import numpy as np
import healpy as hp
matplotlib.use('Agg')
from pathlib import Path
import concurrent.futures
import scipy.optimize as opt
from null_test import rotate
import matplotlib.pyplot as plt
import matplotlib.gridspec as gs
from mpl_toolkits.axes_grid1 import make_axes_locatable
from scipy.stats import chisquare
from scipy.stats import median_absolute_deviation as mad
from scipy.interpolate import make_interp_spline, BSpline
import scipy.optimize as opt
import numpy.polynomial.polynomial as poly

sys.path.append('../sat_ephemeris')
from sat_ids import norad_ids

sys.path.append('../sat_channels')
from sat_channels import time_tree, time_filter

# Custom spectral colormap
sys.path.append('../decode_rf_data')
from rf_data import tile_names
from colormap import spectral
cmap = spectral()


def check_pointing(timestamp, point_0, point_2, point_4, point_41):
    '''Check if timestamp is at pointing 0, 2, 4, 41'''
    if timestamp in point_0:
        point = 0
    elif timestamp in point_2:
        point = 2
    elif timestamp in point_4:
        point = 4
    else:
        point=41
    
    return point

def read_aligned(ali_file=None):
    '''Read aligned data npz file'''
    paired_data = np.load(ali_file, allow_pickle=True)
    
    ref_p   = paired_data['ref_p_aligned']
    tile_p  = paired_data['tile_p_aligned']
    times   = paired_data['time_array']

    return [ref_p, tile_p, times]

def flag_clipped(ref_p, tile_p):
    '''When the input power to the RF Explorer
    exceeds -30 dBm, it is distored. We replace 
    all such values with nans'''

    tile_clip = np.where(tile_p >= rfe_clip)
    ref_p[tile_clip] = np.nan
    tile_p[tile_clip] = np.nan

    return (ref_p, tile_p)

def noise_floor(sat_thresh, noi_thresh, data=None):
    '''Computes the noise floor of the data '''
    
    # compute the standard deviation of data, and use it to identify occupied channels
    σ = np.std(data)
    
    # Any channel with a max power >= σ has a satellite
    sat_cut = sat_thresh*σ
    chans_pow_max = np.amax(data, axis=0)
    
    # Exclude the channels with sats, to only have noise data
    noise_chans = np.where(chans_pow_max < sat_cut)[0]
    noise_data = data[:, noise_chans]
    
    # noise median, noise mad, noise threshold = μ + 3*σ
    μ_noise = np.median(noise_data)
    σ_noise = mad(noise_data, axis=None)
    noise_threshold = μ_noise + noi_thresh*σ_noise
    
    return (data, noise_threshold)


# chisquared minimization to best fit map to data
def fit_gain(map_data=None,fee=None):
    '''Fit the beam model to the measured data using
    chisquared minimization'''
    
    bad_values = np.isnan(map_data)
    map_data = map_data[~bad_values]
    fee = fee[~bad_values]
    
    def chisqfunc(gain):
        model = fee + gain
        chisq = sum((map_data - model)**2)
        return chisq

    x0 = np.array([0])

    result =  opt.minimize(chisqfunc,x0)
    
    return result.x


def fit_test(map_data=None,fee=None):
    '''chi-squared test for fit between
    model and data'''
    
    bad_values = np.isnan(map_data)
    map_data = map_data[~bad_values]
    fee = fee[~bad_values]
    
    map_data = np.asarray(map_data) + 120
    fee = np.asarray(fee) + 120

   
    _, pvalue = chisquare(map_data, f_exp=fee)

    return pvalue

def poly_fit(x, y, order):
    '''Fit polynominal of order to data'''
    
    x = np.asarray(x)
    y = np.asarray(y)

    coefs = poly.polyfit(x, y, order)
    fit = poly.polyval(x, coefs)
    return fit

def plt_fee_fit(t, mwa_fee_pass, mwa_pass_fit, out_dir, point, timestamp, sat):

    pval = fit_test(map_data=mwa_pass_fit, fee=mwa_fee_pass)

    plt.style.use('seaborn')
    
    nice_fonts = {
            # Use LaTeX to write all text
            #"text.usetex": True,
            "font.family": "sans-serif",
            # Use 10pt font in plots, to match 10pt font in document
            "axes.labelsize": 10,
            "font.size": 10,
            # Make the legend/label fonts a little smaller
            "legend.fontsize": 6,
            "xtick.labelsize": 8,
            "ytick.labelsize": 8,
            }
    
    plt.rcParams.update(nice_fonts)
    
    
    fig = plt.figure(figsize=(3.6,2.4))
    ax1 = fig.add_subplot(1, 1, 1)
    
    t = np.array(t)
    mwa_fee_pass = [x for _,x in sorted(zip(t,mwa_fee_pass))]
    mwa_pass_fit = [x for _,x in sorted(zip(t,mwa_pass_fit))]
    t = sorted(t)
    t = (t - t[0])/60

    t_new = np.linspace(t.min(), t.max(), 49)
    spl = make_interp_spline(t, mwa_fee_pass, k=3)  # type: BSpline
    mwa_fee_pass = spl(t_new)
    
    spl = make_interp_spline(t, mwa_pass_fit, k=3)  # type: BSpline
    mwa_pass_fit = spl(t_new)

    ax1.plot(t_new, mwa_fee_pass, color='#7da87b', alpha=0.9, lw=2, label=r"$B_{FEE}$")
    ax1.plot(t_new, mwa_pass_fit, color='#c70039', alpha=0.9, lw=2, label=r"$B_{dist}$")
    #ax1.scatter(t, mwa_fee_pass, color='#7da87b', alpha=0.6, marker=".", s=21, label="FEE model")
    #ax1.scatter(t, mwa_pass_fit, color='#c70039', alpha=0.9, marker=".", s=21, label="RF data")
    ax1.set_ylabel('Power [dBm]')
    ax1.set_ylim([-84,-16])
    ax1.set_xticklabels([])
    
    leg = ax1.legend(loc="upper left", frameon=True, markerscale=2, handlelength=1)
    leg.get_frame().set_facecolor('grey')
    leg.get_frame().set_alpha(0.4)
    for l in leg.legendHandles:
        l.set_alpha(1)
    
    
    delta_p = np.array(mwa_fee_pass) - np.array(mwa_pass_fit)
    
    divider = make_axes_locatable(ax1)
    dax = divider.append_axes("bottom", size="40%", pad=0.10)

    dax.plot(t_new, delta_p, lw=2,alpha=0.7, color='#27296d', label='Residuals')
    leg = dax.legend(loc="lower left", frameon=True, markerscale=2, handlelength=1)
    leg.get_frame().set_facecolor('grey')
    leg.get_frame().set_alpha(0.4)
    for l in leg.legendHandles:
        l.set_alpha(1)
    #dax.scatter(t, delta_p, marker='.', s=21,alpha=0.7, color='#27296d')
    dax.set_xlabel('Times [min]')
    #dax.plot(zen_angle, pow_fit, linewidth=1.4, alpha=0.9, color='#ff8264')
    #dax.set_ylim([-21,12])
    dax.set_yticks([-20,0,10])
    #dax.set_xticklabels([])
    #dax.set_ylim([-5,5])
    dax.set_ylabel(r'$\Delta$P [dBm]')
    
    plt.tight_layout()
    plt.savefig(f'{out_dir}/{timestamp}_{sat}.pdf', bbox_inches='tight')
    plt.close()


def power_ephem(
        ref, tile,
        ali_file,
        chrono_file,
        sat_id,
        sat_chan,
        point,
        pow_thresh,
        timestamp
        ):

    '''Create power, alt, az arrays at constant cadence'''

    # Read .npz aligned file
    ref_p, tile_p, times = read_aligned(ali_file=ali_file)

    # Scale noise floor to zero and determine noise threshold
    ref_p, ref_noise = noise_floor(sat_thresh, noi_thresh, ref_p)
    tile_p, tile_noise = noise_floor(sat_thresh, noi_thresh, tile_p)
    
    with open(chrono_file) as chrono:
        chrono_ephem = json.load(chrono)
    
        norad_list = [chrono_ephem[s]["sat_id"][0] for s in range(len(chrono_ephem))]
        
        norad_index = norad_list.index(sat_id)
        
        norad_ephem = chrono_ephem[norad_index]
            
        rise_ephem  = norad_ephem["time_array"][0] 
        set_ephem   = norad_ephem["time_array"][-1]
        
        intvl = time_filter(rise_ephem, set_ephem, np.asarray(times))
        
        if intvl != None:
            
            w_start, w_stop = intvl

            # Slice [crop] the ref/tile/times arrays to the times of sat pass and extract sat_chan
            ref_c = ref_p[w_start:w_stop+1, sat_chan]
            tile_c = tile_p[w_start:w_stop+1, sat_chan]
            times_c = times[w_start:w_stop+1]
            
            alt = np.asarray(norad_ephem["sat_alt"])
            az  = np.asarray(norad_ephem["sat_az"])

            if ((np.nanmax(ref_c - ref_noise) >= pow_thresh) and (np.nanmax(tile_c - tile_noise) >= pow_thresh)):

                # Apply noise criteria. In the window, where are ref_power and tile power
                # above their respective thresholds?
                if np.where((ref_c >= ref_noise) & (tile_c >= tile_noise))[0].size != 0: 
                    good_ref    = ref_c[np.where((ref_c >= ref_noise) & (tile_c >= tile_noise))[0]]
                    good_tile   = tile_c[np.where((ref_c >= ref_noise) & (tile_c >= tile_noise))[0]]
                    good_alt    = alt[np.where((ref_c >= ref_noise) & (tile_c >= tile_noise))[0]]
                    good_az     = az[np.where((ref_c >= ref_noise) & (tile_c >= tile_noise))[0]]

                    # Clip data
                    #good_ref, good_tile = flag_clipped(good_ref, good_tile)

                    return [good_ref, good_tile, good_alt, good_az, times_c]

                else:
                    return 0
            
            else:
                return 0

        else:
            return 0


def project_tile_healpix(tile_pair):

    #pass_data = []
    #pass_resi = []

    resi_gain = {}
    resi_gain['pass_data'] = []
    resi_gain['pass_resi'] = []

    ref, tile = tile_pair

    pointings = ['0','2','4', '41']
    
    if plots == 'True':
        Path(f'{out_dir}/fit_plots/').mkdir(parents=True, exist_ok=True)

    # Initialize an empty dictionary for tile data
    # The map is list of length 12288 of empty lists to append pixel values to
    # keep track of which satellites contributed which data
    tile_data = {
            'mwa_maps':{p:[[] for pixel in range(hp.nside2npix(nside))] for p in pointings}, 
            'ref_maps':{p:[[] for pixel in range(hp.nside2npix(nside))] for p in pointings}, 
            'tile_maps':{p:[[] for pixel in range(hp.nside2npix(nside))] for p in pointings}, 
            'sat_map':{p:[[] for pixel in range(hp.nside2npix(nside))] for p in pointings},
            'times':{p:[[] for pixel in range(hp.nside2npix(nside))] for p in pointings}
            }
    
    
    # Load reference FEE model
    # Rotate the fee models by -pi/2 to move model from spherical (E=0) to Alt/Az (N=0)
    ref_fee_model = np.load(ref_model, allow_pickle=True)
    
#    # Tile S33YY has it's 9th dipole flagged
#    if tile is 'S33YY':
#        fee_m = np.load(fee_map_flagged, allow_pickle=True)
#    else:
#        fee_m = np.load(fee_map, allow_pickle=True)
    fee_m = np.load(fee_map, allow_pickle=True)
        
    if 'XX' in tile:    
        ref_fee = ref_fee_model['XX']
        rotated_fee = rotate(nside, angle=-(1*np.pi)/2.0,healpix_array=ref_fee)
    else:
        ref_fee = ref_fee_model['YY']
        rotated_fee = rotate(nside, angle=-(1*np.pi)/2.0,healpix_array=ref_fee)
    
    for day in range(len(dates)):

        for window in range(len(date_time[day])):
            timestamp = date_time[day][window]
            
            # Check if at timestamp, reciever was pointed to 0,2,4 gridpointing
            if ((timestamp in point_0) or (timestamp in point_2) or (timestamp in point_4) or (timestamp in point_41)):
            
                # pointing at timestamp
                point = check_pointing(timestamp, point_0, point_2, point_4, point_41)
                 
                if point is 0:

                    if 'XX' in tile:    
                        mwa_fee = fee_m[str(point)][0]
                    else:
                        mwa_fee = fee_m[str(point)][1]

                    ali_file = Path(f'{align_dir}/{dates[day]}/{timestamp}/{ref}_{tile}_{timestamp}_aligned.npz')

                    # check if file exists
                    if ali_file.is_file():

                        # Chrono and map Ephemeris file
                        chrono_file = Path(f'{chrono_dir}/{timestamp}.json')
                        channel_map = f'{map_dir}/{timestamp}.json'

                        with open(chrono_file) as chrono:
                            chrono_ephem = json.load(chrono)
                        
                            if chrono_ephem != []:
                    
                                norad_list = [chrono_ephem[s]["sat_id"][0] for s in range(len(chrono_ephem))]

                                if norad_list != []:

                                    with open(channel_map) as ch_map:
                                        chan_map = json.load(ch_map)
                                    
                                        chan_sat_ids = [int(i) for i in list(chan_map.keys())]
                            
                                        for sat in chan_sat_ids:

                                            chan = chan_map[f'{sat}']

                                            sat_data = power_ephem(
                                                    ref, tile,
                                                    ali_file,
                                                    chrono_file,
                                                    sat,
                                                    chan,
                                                    point,
                                                    pow_thresh,
                                                    timestamp
                                                    )

                                            if sat_data != 0:
                                            
                                                ref_power, tile_power, alt, az, times = sat_data


                                                # Altitude is in deg while az is in radians
                                                # convert alt to radians
                                                # za - zenith angle
                                                alt = np.radians(alt)
                                                za = np.pi/2 - alt
                                                az  = np.asarray(az)


                                                # Now convert to healpix coordinates
                                                #healpix_index = hp.ang2pix(nside,θ, ɸ)
                                                healpix_index = hp.ang2pix(nside, za, az)
                                                
                                                # multiple data points fall within a single healpix pixel
                                                # find the unique pixels
                                                u = np.unique(healpix_index)

                                                ref_pass = np.array([np.nanmean(ref_power[np.where(healpix_index==i)[0]]) for i in u])
                                                tile_pass = np.array([np.nanmean(tile_power[np.where(healpix_index==i)[0]]) for i in u])
                                                times_pass = np.array([np.mean(times[np.where(healpix_index==i)][0]) for i in u])
                                                ref_fee_pass = np.array([rotated_fee[i] for i in u])
                                                mwa_fee_pass = np.array([mwa_fee[i] for i in u])
                                                
                                                #clipped_idx = np.where(tile_pass == np.nan)
                                                #ref_fee_pass[clipped_idx] == np.nan
                                                
                                                #plt.style.use('seaborn')
                                                #nice_fonts = {
                                                #        # Use LaTeX to write all text
                                                #        #"text.usetex": True,
                                                #        "font.family": "sans-serif",
                                                #        # Use 10pt font in plots, to match 10pt font in document
                                                #        "axes.labelsize": 12,
                                                #        "font.size": 12,
                                                #        # Make the legend/label fonts a little smaller
                                                #        "legend.fontsize": 10,
                                                #        "xtick.labelsize": 9,
                                                #        "ytick.labelsize": 9,
                                                #        }
                                                #
                                                #plt.rcParams.update(nice_fonts)
                                                #fig = plt.figure(figsize=(9,5))
                                                #
                                                #ax1 = fig.add_subplot(2, 2, 1)
                                                #ax1.scatter(times_pass-times_pass[0], tile_pass, marker='.', color='#ff6464', label=r'P$_{AUT}$')
                                                #ax1.scatter(times_pass-times_pass[0], ref_pass, marker='.', color='#5e63b6', label=r'P$_{ref}$')
                                                #ax1.scatter(times_pass-times_pass[0], ref_fee_pass, marker='.', color='#17b978', label=r'B$_{ref}$')
                                                #ax1.set_xlabel('Times [s]')
                                                #ax1.set_ylabel('Raw Power [dBm]')
                                                #ax1.legend()
                                                #leg = ax1.legend(loc="upper right", frameon=True, markerscale=2, handlelength=1)
                                                #leg.get_frame().set_facecolor('grey')
                                                #leg.get_frame().set_alpha(0.4)
                                                #for l in leg.legendHandles:
                                                #    l.set_alpha(1)
                                                #
                                                #ax2 = fig.add_subplot(2, 2, 3)
                                                #ax2.scatter(times_pass-times_pass[0], tile_pass-ref_pass+ref_fee_pass, marker='.', color='#ff7c38', label=r'B$_{dist}$ = $\frac{P_{AUT}}{P_{ref}}B_{ref}$')
                                                #ax2.scatter(times_pass-times_pass[0], mwa_fee_pass, marker='.', color='#307672', label=r'B$_{FEE}$')
                                                #ax2.set_xlabel('Times [s]')
                                                #ax2.set_ylabel('Power [dB]')
                                                #ax2.legend()
                                                #leg = ax2.legend(loc="upper right", frameon=True, markerscale=2, handlelength=1)
                                                #leg.get_frame().set_facecolor('grey')
                                                #leg.get_frame().set_alpha(0.4)
                                                #for l in leg.legendHandles:
                                                #    l.set_alpha(1)
                                                #
                                                #ax3 = fig.add_subplot(1, 2, 2)
                                                #mwa_pass = np.array(tile_pass) - np.array(ref_pass) + np.array(ref_fee_pass)
                                                #offset = fit_gain(map_data=tile_pass, fee=mwa_pass)
                                                #mwa_pass = mwa_pass + offset[0]
                                                #offset = fit_gain(map_data=mwa_pass, fee=mwa_fee_pass)
                                                #mwa_fee_pass = mwa_fee_pass + offset
                                                #mwa_pass_fit = mwa_pass
                                                #
                                                #ax3.scatter(times_pass-times_pass[0], mwa_fee_pass, color='#7da87b', alpha=0.9, marker=".", label="FEE slice fit to P$_{AUT}$")
                                                #ax3.scatter(times_pass-times_pass[0], mwa_pass_fit, color='#c70039', alpha=0.9, marker=".", label="B$_{dist}$ fit to P$_{AUT}$")
                                                #ax3.set_ylabel('Power [dBm]')
                                                ##ax1.set_ylim([-84,-16])
                                                ##ax1.set_xticklabels([])
                                            
                                                #leg = ax3.legend(loc="upper right", frameon=True, markerscale=2, handlelength=1)
                                                #leg.get_frame().set_facecolor('grey')
                                                #leg.get_frame().set_alpha(0.4)
                                                #for l in leg.legendHandles:
                                                #    l.set_alpha(1)
                                                #
                                                #
                                                #delta_p = np.array(mwa_fee_pass) - np.array(mwa_pass_fit)
                                                #
                                                #divider = make_axes_locatable(ax3)
                                                #dax = divider.append_axes("bottom", size="40%", pad=0.10)

                                                #dax.scatter(times_pass-times_pass[0], delta_p, marker='.', alpha=0.7, color='#27296d', label='Residuals')
                                                #leg = dax.legend(loc="upper right", frameon=True, markerscale=2, handlelength=1)
                                                #leg.get_frame().set_facecolor('grey')
                                                #leg.get_frame().set_alpha(0.4)
                                                #for l in leg.legendHandles:
                                                #    l.set_alpha(1)
                                                ##dax.scatter(t, delta_p, marker='.', s=21,alpha=0.7, color='#27296d')
                                                #dax.set_xlabel('Times [s]')
                                                ##dax.plot(zen_angle, pow_fit, linewidth=1.4, alpha=0.9, color='#ff8264')
                                                ##dax.set_ylim([-21,12])
                                                #dax.set_yticks([-20,0,10])
                                                ##dax.set_xticklabels([])
                                                ##dax.set_ylim([-5,5])
                                                #dax.set_ylabel(r'$\Delta$P [dB]')
                                                #
                                                #
                                                #
                                                #plt.tight_layout()
                                                #plt.savefig(f'{out_dir}/{timestamp}_{sat}.png')
                                                #plt.close()

                                                # magic here
                                                # the beam shape finally emerges
                                                
                                                mwa_pass = np.array(tile_pass) - np.array(ref_pass) + np.array(ref_fee_pass)

                                                # fit the mwa_pass data to the tile_pass power level. RFE clipping occurs at the tile_pass power levels                     
                                                peak_filter = np.where(tile_pass >= -40)
                                                offset = fit_gain(map_data=tile_pass[peak_filter], fee=mwa_pass[peak_filter])
                                                mwa_pass = mwa_pass + offset[0]


                                                # fit the power level of the pass to the mwa_fee model using a single gain value
                                                #offset = fit_gain(map_data=mwa_pass, fee=mwa_fee_pass)
                                                #mwa_pass_fit = mwa_pass - offset[0]
                                                
                                                # here, we fit the mwa_fee_pass down to the tile_pass level
                                                # mask out distorted poetion >= -40 dBm
                                                dis_filter = np.where(mwa_pass <= -40) 
                                                mwa_pass_fil = mwa_pass[dis_filter]
                                                mwa_fee_pass_fil = mwa_fee_pass[dis_filter]
                                                # ignore fee nulls in fit
                                                null_filter = np.where(mwa_fee_pass_fil >= -60)

                                                offset = fit_gain(map_data=mwa_pass_fil[null_filter], fee=mwa_fee_pass_fil[null_filter])
                                                mwa_fee_pass = mwa_fee_pass + offset
                                                mwa_pass_fit = mwa_pass
                                                
                                                # determine how well the data fits the model with chi-square  
                                                pval = fit_test(map_data=mwa_pass_fit, fee=mwa_fee_pass)


                                                
                                                
                                                # a goodness of fit threshold
                                                if pval >= 0.9:
                                                
                                                    ##peak = np.where(mwa_fee_pass >= -50)
                                                    #peak = np.where(mwa_pass_fit >= -40)
                                                    #mwa_fee_pass = mwa_fee_pass[peak]
                                                    #mwa_pass_fit = mwa_pass_fit[peak]
                                                    #times_pass = times_pass[peak]
                                                    #hp_indices = u[peak]

                                                    #residuals = mwa_fee_pass - mwa_pass_fit
                                                    #filtr = np.where(residuals > -1)
                                                    #mwa_fee_pass = mwa_fee_pass[filtr]
                                                    #mwa_pass_fit = mwa_pass_fit[filtr]
                                                    #times_pass = times_pass[filtr]
                                                    #resi = mwa_fee_pass - mwa_pass_fit
                                                   
                                                    # consider residuals in the primary lobe approximately
                                                    # an hp index of 111 approx corresponds to a zenith angle of 10 degrees
                                                    hp_10_deg = 111
                                                   
                                                    if np.amin(u) <= hp_10_deg:
                                                        #filtr = np.where(u <= hp_10_deg)
                                                        #mwa_fee_pass = mwa_fee_pass[filtr]
                                                        #mwa_pass_fit = mwa_pass_fit[filtr]
                                                        #times_pass = times_pass[filtr]
                                                    
                                                        resi = mwa_fee_pass - mwa_pass_fit
                                                        resi_gain['pass_data'].extend(mwa_pass_fit)
                                                        resi_gain['pass_resi'].extend(resi)


                                                        if plots == 'True':
                                                            if mwa_fee_pass.size !=0:
                                                                if np.amax(mwa_fee_pass) >= -30:
                                                                    plt_fee_fit(
                                                                            times_pass,
                                                                            mwa_fee_pass, 
                                                                            mwa_pass_fit, 
                                                                            f'{out_dir}/fit_plots/', 
                                                                            point, timestamp, sat)
   
    with open(f'{out_dir}/{tile}_{ref}_gain_fit.json', 'w') as outfile:
        json.dump(resi_gain, outfile, indent=4)    
#    plt.style.use('seaborn')
#    plt.scatter(resi_gain['pass_data'], resi_gain['pass_resi'], marker='.', alpha=0.7, color='seagreen')
#    
#    #pass_data = [x for x,_ in sorted(zip(pass_data,pass_resi))]
#    #pass_resi = [x for _,x in sorted(zip(pass_data,pass_resi))]
#    fit = poly_fit(resi_gain['pass_data'], resi_gain['pass_resi'], 3)
#    plt.plot(sorted(resi_gain['pass_data'], reverse=True), sorted(fit, reverse=True), color='crimson')
#    plt.xlabel('Observed power [dB]')
#    plt.ylabel('Residuals [dB]')
#    plt.tight_layout()
#    plt.savefig(f'{out_dir}/{tile}_{ref}_gain_fit.png')


if __name__=='__main__':

    parser = argparse.ArgumentParser(description="""
        Project a sat pass onto a healpix map using ephemeris data
        """)
    
    parser.add_argument(
            '--align_dir', metavar='\b',default='../../outputs/align_data/',
            help='Dir where agligned data date is saved. Default:../../outputs/align_data')

    parser.add_argument(
            '--out_dir', metavar='\b', default='./../../outputs/tile_maps/tile_maps_raw/',
            help='Output directory. Default=./../../outputs/tile_maps/tile_maps_raw/')
    
    parser.add_argument(
            '--obs_point', metavar='\b', default='../../outputs/beam_pointings/obs_pointings.json',
            help='Observation pointing lists. Default=../../outputs/beam_pointings/obs_pointings.json')

    parser.add_argument(
            '--chrono_dir', metavar='\b', default='./../../outputs/sat_ephemeris/chrono_json',
            help='Output directory. Default=./../../outputs/sat_ephemeris/chrono_json/')
    
    parser.add_argument(
            '--map_dir', metavar='\b', default='../../outputs/sat_channels/window_maps/',
            help='Satellite channel map. Default=../../outputs/sat_channels/window_maps/')

    parser.add_argument('--ref_model', metavar='\b', default='../../outputs/reproject_ref/ref_dipole_models.npz',
            help='Healpix reference FEE model file. default=../../outputs/reproject_ref/ref_dipole_models.npz')
    
    parser.add_argument('--fee_map', metavar='\b', default='../../outputs/tile_maps/FEE_maps/mwa_fee_beam.npz',
            help='Healpix FEE map of mwa tile. default=../../outputs/tile_maps/FEE_maps/mwa_fee_beam.npz')
    
    parser.add_argument('--fee_map_flagged', metavar='\b', default='../../outputs/tile_maps/FEE_maps/mwa_fee_beam_9_flagged.npz',
            help='Healpix FEE map of mwa tile. default=../../outputs/tile_maps/FEE_maps/mwa_fee_beam_9_flagged.npz')
    
    parser.add_argument('--fit_thresh', metavar='\b', default=0.9, help='Goodness of fit threshold. 0.9 seems to only reject obvious outliers')
    parser.add_argument('--noi_thresh', metavar='\b', type=int, default=3,help='Noise Threshold: Multiples of MAD. Default=3.')
    parser.add_argument('--sat_thresh', metavar='\b', type=int, default=1,help='σ threshold to detect sats Default=1.')
    parser.add_argument('--pow_thresh', metavar='\b', type=int, default=5,help='Power Threshold to detect sats. Default=10 dB.')
    parser.add_argument('--rfe_clip', metavar='\b', type=int, default=-30, help='RF Explorer clipping level. Default: -30dBm.')
    parser.add_argument('--nside', metavar='\b', type=int,  default=32,help='Healpix Nside. Default = 32')
    parser.add_argument('--plots', metavar='\b', default=False,help='If True, create a gazzillion plots for each sat pass. Default = False')
    parser.add_argument('--start_date', metavar='\b', help='Date from which to start aligning data. Ex: 2019-10-10')
    parser.add_argument('--stop_date', metavar='\b', help='Date until which to align data. Ex: 2019-10-11')
    
    args = parser.parse_args()
    
    start_date      = args.start_date
    stop_date       = args.stop_date
    obs_point       = args.obs_point
    noi_thresh      = args.noi_thresh
    sat_thresh      = args.sat_thresh
    pow_thresh      = args.pow_thresh
    fit_thresh      = args.fit_thresh
    rfe_clip        = args.rfe_clip
    nside           = args.nside
    plots           = args.plots
    ref_model       = args.ref_model
    fee_map         = args.fee_map
    fee_map_flagged = args.fee_map_flagged
    
    align_dir       = Path(args.align_dir)
    chrono_dir      = Path(args.chrono_dir)
    out_dir         = Path(args.out_dir)
    map_dir         = Path(args.map_dir)

    # Tile names
    refs    = tile_names()[:4]
    tiles   = tile_names()[4:]
    
    # All relevant tile pairs
    tile_pairs = []
    for ref in refs:
        if 'XX' in ref:
            for tile in [t for t in tiles if 'XX' in t]:
                tile_pairs.append([ref, tile])
        else:
            for tile in [t for t in tiles if 'YY' in t]:
                tile_pairs.append([ref,tile])

    # Read observation pointing list
    with open(obs_point) as point:
        obs_p = json.load(point)
        point_0  = obs_p['point_0'] 
        point_2  = obs_p['point_2'] 
        point_4  = obs_p['point_4']
        point_41 = obs_p['point_41']


    # dates: list of days
    # date_time = list of 30 min observation windows
    dates, date_time = time_tree(start_date, stop_date)

    # Save logs 
    Path(out_dir).mkdir(parents=True, exist_ok=True)
#    sys.stdout = open(f'{out_dir}/logs_{start_date}_{stop_date}.txt', 'a')
   
#    project_tile_healpix(tile_pairs[0])
#    project_tile_healpix(['rf0YY', 'S33YY'])
        
    # Parallization magic happens here
    with concurrent.futures.ProcessPoolExecutor() as executor:
        results = executor.map(project_tile_healpix, tile_pairs)





import sys
import json
import argparse
import matplotlib
import numpy as np
matplotlib.use('Agg')
from pathlib import Path
import scipy.optimize as opt
import matplotlib.pyplot as plt
import matplotlib.gridspec as gs
from scipy.stats import median_absolute_deviation as mad
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy.polynomial.polynomial as poly
from scipy.stats import binned_statistic
import scipy.optimize as opt

sys.path.append('../decode_rf_data')
from colormap import spectral
cmap = spectral()

gain_files = [item for item in Path('../../outputs/paper_plots/gain_fit/').glob('*.json')]
names = [i.name.split('.')[0] for i in gain_files]

pass_data = []
pass_resi = []

for n,f in enumerate(gain_files):
    with open(f, 'r') as data:
        rfe = json.load(data)
        pass_data.extend(rfe['pass_data'])
        pass_resi.extend(rfe['pass_resi'])
        
nice_fonts = {
        # Use LaTeX to write all text
        #"text.usetex": True,
        "font.family": "sans-serif",
        # Use 10pt font in plots, to match 10pt font in document
        "axes.labelsize": 10,
        "font.size": 10,
        # Make the legend/label fonts a little smaller
        "legend.fontsize": 6,
        "xtick.labelsize": 8,
        "ytick.labelsize": 8,
        }

plt.rcParams.update(nice_fonts)

fig = plt.figure(figsize=(3.6,2.4))
#fig = plt.figure()

plt.hexbin(pass_data, pass_resi, gridsize=77,cmap=cmap, alpha=0.99, zorder=0)

pass_data = np.array(pass_data)
pass_resi = np.array(pass_resi)
filtr = np.where(np.logical_and(pass_data <= -20, pass_data >=-70))
pass_data = pass_data[filtr]
pass_resi = pass_resi[filtr]

# Median of binned data
bin_med, bin_edges, binnumber = binned_statistic(pass_data, pass_resi, statistic='median', bins=20)
bin_width = (bin_edges[1] - bin_edges[0])
bin_centers = bin_edges[1:] - bin_width/2

# Now look at data between -50, -30, where gains vary
filtr = np.where(np.logical_and(pass_data <= -28, pass_data >=-52))
pass_data = pass_data[filtr]
pass_resi = pass_resi[filtr]

z = np.polyfit(pass_data, pass_resi, 1)
f = np.poly1d(z)

x_f = [f.roots[0], -45, -40, -35, -30, -25, -20]
y_f = f(x_f)

#plt.plot(x_f, y_f, color='w', lw=2.1, marker='X', markeredgecolor='k', markeredgewidth=0.7, markersize=2.1, alpha=1, label='Gain fit', zorder=1)
plt.plot(x_f, y_f, color='w', lw=1.8, alpha=1, label='Gain fit', zorder=1)
plt.scatter(bin_centers, bin_med, marker='h', s=14,  facecolors='w',lw=0.7, edgecolors='k', alpha=1, label='Median residuals', zorder=2)
plt.scatter(f.roots[0], 0, marker='o', s=9,  facecolors='w', alpha=1)

leg = plt.legend(loc="lower right", frameon=True,framealpha=0.7, markerscale=0.9, handlelength=1)
leg.get_frame().set_facecolor('#cccccc')
for l in leg.legendHandles:
    l.set_alpha(1)


plt.xlabel('Observed power [dBm]')
plt.ylabel('Residuals power [dB]')
plt.xlim([-70,-20])
plt.ylim([-10,15])
plt.tick_params(axis='both', length = 0)
plt.grid(color='#cccccc', alpha=0.21, lw=1)
plt.box(None)
plt.tight_layout()
plt.savefig(f'../../outputs/paper_plots/rfe_gain_fit.pdf', bbox_inches='tight')




# A dictionary of satellites which emit between 137.150 and 138.550 MHz
# ORBCOMM Communication satellites, NOAA & METEOR Weather satellites
# Satellite Name : NORAD CATALOGUE ID

norad_ids = {
        'ORBCOMM-X': 21576,
        'ORBCOMM FM 1': 23545,
        'ORBCOMM FM 2': 23546,
        'ORBCOMM FM 8': 25112,
        'ORBCOMM FM 10': 25113,
        'ORBCOMM FM 11': 25114,
        'ORBCOMM FM 12': 25115,
        'ORBCOMM FM 9': 25116,
        'ORBCOMM FM 5': 25117,
        'ORBCOMM FM 6': 25118,
        'ORBCOMM FM 7': 25119,
        'ORBCOMM FM 3': 25158,
        'ORBCOMM FM 4': 25159,
        'ORBCOMM FM 17': 25413,
        'ORBCOMM FM 18': 25414,
        'ORBCOMM FM 19': 25415,
        'ORBCOMM FM 20': 25416,
        'ORBCOMM FM 16': 25417,
        'ORBCOMM FM 15': 25418,
        'ORBCOMM FM 14': 25419,
        'ORBCOMM FM 13': 25420,
        'ORBCOMM FM 21': 25475,
        'ORBCOMM FM 22': 25476,
        'ORBCOMM FM 23': 25477,
        'ORBCOMM FM 24': 25478,
        'ORBCOMM FM 25': 25479,
        'ORBCOMM FM 26': 25480,
        'ORBCOMM FM 27': 25481,
        'ORBCOMM FM 28': 25482,
        'ORBCOMM FM 30': 25980,
        'ORBCOMM FM 31': 25981,
        'ORBCOMM FM 32': 25982,
        'ORBCOMM FM 33': 25983,
        'ORBCOMM FM 36': 25984,
        'ORBCOMM FM 35': 25985,
        'ORBCOMM FM 34': 25986,
        'ORBCOMM FM 38': 33060,
        'ORBCOMM FM 41': 33061,
        'ORBCOMM FM 29': 33062,
        'ORBCOMM FM 39': 33063,
        'ORBCOMM FM 37': 33064,
        'ORBCOMM FM 40': 33065,
        'ORBCOMM FM 109': 40086,
        'ORBCOMM FM 107': 40087,
        'ORBCOMM FM 106': 40088,
        'ORBCOMM FM 111': 40089,
        'ORBCOMM FM 104': 40090,
        'ORBCOMM FM 103': 40091,
        'ORBCOMM FM 114': 41179,
        'ORBCOMM FM 119': 41180,
        'ORBCOMM FM 105': 41181,
        'ORBCOMM FM 110': 41182,
        'ORBCOMM FM 118': 41183,
        'ORBCOMM FM 112': 41184,
        'ORBCOMM FM 113': 41185,
        'ORBCOMM FM 115': 41186,
        'ORBCOMM FM 108': 41187,
        'ORBCOMM FM 117': 41188,
        'ORBCOMM FM 116': 41189,
        'ORBCOMM FM 16 DEB 1': 44018,
        'ORBCOMM FM 16 DEB 2': 44019,
        'ORBCOMM FM 16 DEB 3': 44020, # [NO CURRENT TLEs FOUND!]
        'ORBCOMM FM 16 DEB 4': 44022,
        'ORBCOMM FM 16 DEB 5': 44023,
        'ORBCOMM FM 16 DEB 6': 44024,
        'ORBCOMM FM 16 DEB 7': 44025, # [NO CURRENT TLEs FOUND!]
        'ORBCOMM FM 16 DEB 8': 44026,
        'ORBCOMM FM 16 DEB 9': 44027,
        'ORBCOMM FM 16 DEB 10': 44028,
        'NOAA 18': 28654,
        'NOAA 15': 25338,
        'Meteor M2': 40069,
        'Meteor M2-2': 44387
        }
    
if __name__ == '__main__':
    for ids in norad_ids.values():
        print(ids)
